/*
    `yy' - A lightshow-like program that graphically represents sound.
    Copyright (C) 1998  Nikos Gorogiannis <kgor@softlab.ece.ntua.gr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
    See file `copyright' for details.
*/

#ifndef DEF_SHARED_HH
#define DEF_SHARED_HH

#include <iostream.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <math.h>

typedef unsigned char Pixel;
// video-ram is 8-bit. Some of these are used for brightness...
#define BRIGHT_BITS 6
// ...and the rest for 'color' selection
#define PALETTE_BITS (8 - BRIGHT_BITS)

#define MAX_BRIGHT ((1<<BRIGHT_BITS)-1)
#define MAX_PALETTE ((1<<PALETTE_BITS)-1)

#define BRIGHT_MASK MAX_BRIGHT
#define PALETTE_MASK (MAX_PALETTE<<BRIGHT_BITS)

#define brightness(c) (BRIGHT_MASK & (c))
#define palette(c) (PALETTE_MASK & (c))
#define pcolor(p,c) ((Pixel) ((p) | (c)))


// for shared-variable palette
class RGBcolor
{
public:
        double red;
        double green;
        double blue;

        inline RGBcolor(const double r = 0.0, 
		const double g = 0.0, 
		const double b = 0.0);
};

inline RGBcolor::RGBcolor(const double r, const double g, const double b) :
	red(r), green(g), blue(b) {}


class HSVcolor
{
public:
        double hue;
        double saturation;
        double value;

        inline HSVcolor(const double h = 0.0, 
		const double s = 0.0, 
		const double v = 0.0);

        inline operator RGBcolor(void) const volatile;

};

inline HSVcolor::HSVcolor(const double h, const double s, const double v) :
	hue(h), saturation(s), value(v) {}

inline HSVcolor::operator RGBcolor(void) const
 volatile
{
	if(saturation==0.0)
	    return RGBcolor(value,value,value);

	const double h = hue/60.0;
	const int i = (int) floor(h);
	const double f = h - i;
	const double p = value*(1.0 - saturation);
	const double q = value*(1.0 - saturation*f);
	const double t = value*(1.0 - saturation*(1.0 - f));

	switch(i)
	  {
	  case 0:
		  return RGBcolor(value,t,p);
	  case 1:
		  return RGBcolor(q,value,p);
	  case 2:
		  return RGBcolor(p,value,t);
	  case 3:
		  return RGBcolor(p,q,value);
	  case 4:
		  return RGBcolor(t,p,value);
	  default:
		  return RGBcolor(value,p,q);
	  }
}


#define HSV_RED_LOW    0.0
#define HSV_YELLOW    60.0
#define HSV_GREEN    120.0
#define HSV_CYAN     180.0
#define HSV_BLUE     240.0
#define HSV_MAGENTA  300.0
#define HSV_RED_HIGH 360.0


const double pi2 = atan(1.0)*4.0/2.0;

// 'states' of screener - either stable or in transit
#define STATE_STABLE 0
#define STATE_TRANSIT 1

// shared variables structure
typedef volatile struct 
{
	// requested attenuation in terms of (0.0 ... 1.0)/decisecond
	double rAttenuation;
	
	// percent of aging int terms of (0.0-1.0)/decisecond
	double rAging;

	// sort of 'stereo enhancement'
	double aperture;
	
	// spots appear for mean intensities over this
	double threshold;
	
	// termination flag for child processes
	int terminate;
	
	// when stable, index of current transformation
	// when in transit, index of previous transformation
	unsigned transfIndex1;
	// used only when in transit, index of new transformation
	unsigned transfIndex2;
	// state, stable or in transit
	unsigned state;
	// used only when in transit, 0.0=begining of transition, 1.0=end
	double stateFraction;
	
	// MAX_PALETTE+1 colors exist and others are deduced by interpolation
	HSVcolor palette[MAX_PALETTE+1];
	
 	int paletteChange;
	
	unsigned colorMixerIndex;
	
	unsigned iconMinSize;
	unsigned iconMaxSize;
} SharedVars;

// shared memory pointer for shared variables
extern SharedVars *sharedVars;

extern int requestedMode;

// for debugging macros chkRetVal, chkVal, chk, dbg
#define DBG

// dbg macro just prints message
#ifdef DBG
#define dbg(msg) cerr << msg
#else
#define dbg(msg) /**/
#endif

// macro prints diagnostic info, message and quits
#define error(msg) \
do { cerr << \
"\nFile= " __FILE__ << \
"\nLine= " << __LINE__  << \
"\nMessage= " msg << \
"\nError= " << strerror(errno) << '\n'; \
if(sharedVars!=NULL) sharedVars->terminate=1; \
exit(1); \
} while(0)

//"\nFunction= " __FUNCTION__ << 

// macro checks var==errval and if true quits
#ifdef DBG
#define chkVal(var, errval) \
do { \
if((var)==(errval)) \
error("Variable:" #var " is " #errval); \
} while(0)
#else
#define chkVal(var, errval) /* */
#endif


// macro runs func and checks retval
#ifdef DBG
#define chkRetVal(func, errval) \
do { \
if((func)==(errval)) \
error("Function:" #func " returned " #errval); \
} while(0)
#else
#define chkRetVal(func, errval) (func)
#endif

// macro runs func and checks retval==-1
#define chk(func) chkRetVal(func, -1)

// macro used to report 'user' errors- ie errors not concerning int. cons.
#define userError(msg)\
do { \
cerr << msg << '\n'; \
if(sharedVars!=NULL) sharedVars->terminate=1; \
exit(1); \
} while(0)

#endif
