/*
    `yy' - A lightshow-like program that graphically represents sound.
    Copyright (C) 1998  Nikos Gorogiannis <kgor@softlab.ece.ntua.gr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
    See file `copyright' for details.
*/

//#include <stdio.h>
//#include <unistd.h>
//#include <fcntl.h>
#include <stdlib.h>
#include <sys/shm.h>
#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>

#include "def_shared.hh"
#include "def_settings.hh"
#include "sema.hh"
#include "icons.hh"

extern void recorder(void);
extern void screener(void);
extern int yyparse(void);

extern FILE * yyin;

int requestedMode=0;

class Process
{
public:
	volatile int pid;
	
	Process(void);
	void spawn(void (*function)());
	void terminate(void);
	void wait(void);
};

Process::Process(void) : pid(-1) {}

void Process::terminate(void)
{
	if(pid<=0)
		return;

	if( kill(pid, SIGTERM) != 0 )
		kill(pid, SIGINT);
	
	pid = -1;
}

void Process::spawn(void (*function)())
{
	pid = fork();
	
	if(pid==0)
	{
		function();
		exit(0);
	}
}

void Process::wait(void)
{
	if(pid==-1)
		return;
	
	while( (waitpid(pid, NULL, 0)==-1) && (errno==EINTR) );
}

Process child[2];
#define SCREENER 0
#define RECORDER 1


// shared memory recording buffer
volatile Sample *buffer;

// shared memory for shared variables
SharedVars *sharedVars = NULL;


class SharedMem
{
public:
	SharedMem(void);
	~SharedMem(void);
	void * create(int size);
	void destroy(void);
	
private:
	void * hole;
	int shmid;
};

SharedMem::SharedMem(void)
{
	hole = NULL;
	shmid = -1;
}

SharedMem::~SharedMem(void)
{
	if(hole != NULL)
	  free(hole);
}

void * SharedMem::create(int size)
{
	// svgalib does not like autoallocated shared segments
	// so allocate a /hole/ which will be used for remaping
	// shared memory. make it page-aligned because ipc-docs *suck*.
	// make it page-rounded because ipc-docs suck too.
	// for libc6 i had to look into include files to find pvalloc. great.
	// pvalloc is valloc with rounded-up size to pagesize()
	hole = (void *) pvalloc(size);
	chkVal(hole, NULL);
	
	// get shared memory
	shmid = shmget(IPC_PRIVATE, size, IPC_CREAT | IPC_EXCL | 0660);
	chkVal(shmid, -1);

	// remap it at sharedVars
	void * tmp = shmat(shmid, hole, SHM_REMAP);
        chkVal((int) tmp, -1);
	
	return tmp;
}

void SharedMem::destroy(void)
{
	if(shmid!=-1)
		shmctl(shmid, IPC_RMID, NULL);
	// no error control since this could be invoked by an error-
	// -terminated function
}

// handler asynchronously updates info in Process classes regarding death...
void sigchld_handler(int signum)
{
	while(1)
	{
		int pid = waitpid(WAIT_ANY, NULL, WNOHANG);

		if(pid<=0)
			break;
		
		if( child[SCREENER].pid == pid)
		{
			child[SCREENER].pid = -1;
			//cerr << "\nScreener is dead";
		}
		
		if( child[RECORDER].pid == pid)
		{
			child[RECORDER].pid = -1;
			//cerr << "\nRecorder is dead";
		}
	};
}


void sigint_handler(int signum)
{
	// just call exit, in order to call onTermination
	exit(1);
}


SharedMem sh1, sh2;
Semaphore sema;

void onTermination(void)
{
	// notify children
	if(sharedVars != NULL)
		sharedVars->terminate = 1;

	// wait for childrens' exit
	child[SCREENER].wait();
	child[RECORDER].wait();
	
	// now that noone uses shared memory/semaphore, destroy it
	sh1.destroy();
	sh2.destroy();
	sema.destroy();
}

int main(int argc, char **argv)
{
	// install signal handler for ^C
	signal( SIGINT, sigint_handler );
	
	if(argc!=3)
	    userError("Invalid number of arguments.\nUsage: yy filename videoMode");
	    
	// redirect parser input
	yyin = fopen(argv[1], "r");
	chkVal(yyin, NULL);
	
	requestedMode = atoi(argv[2]);
	
	// create shared memory segments
	sharedVars = (SharedVars *) sh1.create( sizeof(SharedVars) + 100);
	buffer = (volatile Sample *) sh2.create(BUF_SIZE);
	
	// create semaphore
	sema.create();

        // fill in necessary vars with default values
        sharedVars->rAttenuation = 0.01;
	sharedVars->rAging = 0.99;
        sharedVars->threshold = 0.5;
        sharedVars->terminate = 0;
	sharedVars->transfIndex1 = 0;
	sharedVars->state = STATE_STABLE;
	sharedVars->paletteChange = 1;
	sharedVars->aperture = 15.0;
	sharedVars->colorMixerIndex = 1;
	sharedVars->iconMinSize = 0;
	sharedVars->iconMaxSize = ICONS;

	// install handler for child's death
	signal( SIGCHLD, sigchld_handler );

	// fork other processes
	child[SCREENER].spawn( screener );
	child[RECORDER].spawn( recorder );
	
	// on exit clean up
	atexit(onTermination);

	// run parser
	yyparse();

	return 0;
}
