/*
    `yy' - A lightshow-like program that graphically represents sound.
    Copyright (C) 1998  Nikos Gorogiannis <kgor@softlab.ece.ntua.gr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
    See file `copyright' for details.
*/

#include <iostream.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <rfftw.h>

#include "def_shared.hh"
#include "def_settings.hh"

#include "trans.hh"
#include "framebuffer.hh"
#include "sema.hh"

//#include "function.hh"
//static const IntFunction iLog(log10, 1.0, 1e7, 10000);
//static const IntFunction iAtan(atan, -100.0, 100.0, 10000);
//static const IntFunction iSqrt(sqrt, 0.0, 1e13, 10000);


rfftw_plan plan;

fftw_real in_left[SAMPLES],
	in_right[SAMPLES],
	out_left[SAMPLES],
	out_right[SAMPLES];

double measmax[SAMPLES/2],
	measMean[SAMPLES/2],
	left[SAMPLES/2],
	right[SAMPLES/2],
	phase_diff[SAMPLES/2];

double attenuation = 0.0, aging = 0.5;

FrameBuffer fb;

CoordFunc mixX1, mixX2, mixY1, mixY2;

extern Semaphore sema;

const double log2 = log(2);

void updatePalette(void)
{
	for(unsigned grad=0; grad<=MAX_PALETTE; grad++)
		for(unsigned i=0; i<=MAX_BRIGHT; i++)
		{
			HSVcolor ic = sharedVars->palette[grad];
			ic.value *= (double)i/(double)MAX_BRIGHT;
			RGBcolor irgb = (RGBcolor) ic;
			
			fb.setcolor(pcolor(grad<<BRIGHT_BITS, i), 
					 (int) (irgb.red*63.0), 
					 (int) (irgb.green*63.0), 
					 (int) (irgb.blue*63.0));
		}
}


unsigned mixX(const TransInput * ti)
{
	return (unsigned)
		(sharedVars->stateFraction * mixX2(ti) +
		(1.0-sharedVars->stateFraction) * mixX1(ti));
}

unsigned mixY(const TransInput * ti)
{
	return (unsigned)
		(sharedVars->stateFraction * mixY2(ti) +
		(1.0-sharedVars->stateFraction) * mixY1(ti));
}

inline double norm(const double re, const double im)
{
	double a = re*re + im*im;

	if(a < 4.0)
		a = 4.0;

	// i find that log() is much closer to human perception
	// of sound intensities (like dBs)
	return 0.5*log(a) - log2;
}

inline double phase(const double re, const double im)
{
	if(fabs(re)<1e-6)
		return 0.0;
	else
		return atan( im/re );
}

void innerLoop(void)
{
	CoordFunc X, Y;
	TransInput ti;
	
	struct timeval tv1, tv2;
	unsigned usecs = 75000;
	
	gettimeofday(&tv1, NULL);

	while(1)
	{
		sema.wait();
		// copy input channels to 'in' array
		for(register unsigned i=0,j=0; i<SAMPLES; i++)
		{
			in_left[i] = buffer[j];
			j++;
			in_right[i] = buffer[j];
			j++;
		}
		sema.signal();

		// do fft
		rfftw_one(plan, in_left, out_left);
		rfftw_one(plan, in_right, out_right);


//		left[0]  = norm(out_left[0], 0.0);
//		right[0] = norm(out_right[0], 0.0);
//		phase_diff[0] = 0.0;

		for(register unsigned i=1; i<(SAMPLES+1)/2; i++)
		{
			left[i] = norm(out_left[i], out_left[SAMPLES-i]);
			right[i] = norm(out_right[i], out_right[SAMPLES-i]);

			phase_diff[i] = (phase(out_right[i], out_right[SAMPLES-i]) - 
				phase(out_left[i], out_left[SAMPLES-i]) )/pi2;
		}	

		left[SAMPLES/2]  = norm(out_left[SAMPLES/2], 0.0);
		right[SAMPLES/2] = norm(out_right[SAMPLES/2], 0.0);
		phase_diff[SAMPLES/2] = 0.0;


		// calculation of maximums and of their attenuation
		// start at 1 instead of 0 to skip DC
		for(register unsigned i=1; i<SAMPLES/2; i++)
		{
			const double tmpd = ( left[i] + right[i] )/2.0;

			if( tmpd > measmax[i] )
				measmax[i] = tmpd;
			else
				measmax[i] -=
					(measmax[i]-tmpd) * attenuation;

			measMean[i] = tmpd/measmax[i];
		}

		

		// 1 or 2 transformations?
		if(sharedVars->state==STATE_STABLE)
		{
			X = transformations[sharedVars->transfIndex1].xFunc;
			Y = transformations[sharedVars->transfIndex1].yFunc;
		}
		else
		{
			X = mixX;
			Y = mixY;
			mixX1 = transformations[sharedVars->transfIndex1].xFunc;
			mixY1 = transformations[sharedVars->transfIndex1].yFunc;
			
			mixX2 = transformations[sharedVars->transfIndex2].xFunc;
			mixY2 = transformations[sharedVars->transfIndex2].yFunc;

		}
		
		// 'age' framebuffer
		fb.age(aging);
		
		// display
		for(unsigned i=1; i<SAMPLES/2; i++)
		{
			ti.freqIndex = i;
			
			ti.measMean = measMean[i];
			
			// cut all samples below threshold
			if(ti.measMean<=sharedVars->threshold)
				continue;

			// calculate brightness
			const Pixel c = (Pixel)
				(((double)MAX_BRIGHT)*
				(ti.measMean-sharedVars->threshold)/
				(1.0-sharedVars->threshold));
			
			// don't mind about black samples
			if(c==0)
				continue;
			
			// simple heuristic function where predominance of
			// one channel gives the angle of the sound source
			const double phi = (double)(right[i] - left[i]) /
				(double)(right[i] + left[i]);
			
			// after some statistics, i saw that the forementioned
			// heuristic gives an exponential distribution
			// so, passing it through the 'same' exponential
			// function, it should be transformed to a random
			// variable with a flat distribution
			if(phi<0.0)
				ti.phi = -(1.0-exp(-sharedVars->aperture*(-1.0)*phi));
			else
				ti.phi = 1.0 - exp(-sharedVars->aperture*phi);

			const unsigned char p = (unsigned char)
				(MAX_PALETTE*fabs(phase_diff[i]));
			
			ti.phaseDiff = phase_diff[i];
					
			fb.icon(
				fb.width*X(&ti)/MAXX,
				fb.height*Y(&ti)/MAXY,
				sharedVars->iconMinSize+rand()%(sharedVars->iconMaxSize-sharedVars->iconMinSize),
				p<<BRIGHT_BITS,
				c);
		}

		// update screen
		fb.copyToScreen();

		if(sharedVars->paletteChange)
		{
			updatePalette();
			sharedVars->paletteChange = 0;
		}

		if(sharedVars->terminate)
		{
			cerr << "\noutput=" << 1e6/(double)usecs << " frags/sec ";
			cerr << "or " << SAMPLES*1e6/(double)usecs << " samples/sec.";
			cerr << "\nusecs=" << usecs << '\n';
			break;
		}

		gettimeofday(&tv2, NULL);
		if( (tv1.tv_sec!=tv2.tv_sec) || (tv1.tv_usec!=tv2.tv_sec) )
		{
			usecs = (tv2.tv_sec - tv1.tv_sec)*1000000 +
				(tv2.tv_usec - tv1.tv_usec);
		
			attenuation = pow(sharedVars->rAttenuation, 
					(double)usecs/1e5);
				
			aging = pow(sharedVars->rAging,
				(double)usecs/1e5);
			
			tv1 = tv2;
		}
	}
}

void screener(void)
{
	// framebuffer-buffer
	fb.create();

	// setup graphics mode
	updatePalette();
	initTransformations();
	
	// fourier plan
	plan = rfftw_create_plan(SAMPLES, FFTW_FORWARD, FFTW_ESTIMATE);
	chkVal(plan, NULL);
	
	for(unsigned i=0; i<SAMPLES/2; i++)
		measmax[i] = 0.0;
	
	innerLoop();
	
	rfftw_destroy_plan(plan);

	fb.destroy();	
}
